import os
import json
import glob
import numpy as np

import traceback
import threading
from queue import Queue
from threading import Thread
import sys
import random

import alfworld.agents
from alfworld.agents.utils.misc import Demangler, get_templated_task_desc, add_task_to_grammar
from alfworld.env.thor_env import ThorEnv
from alfworld.agents.expert import HandCodedThorAgent, HandCodedAgentTimeout
from alfworld.agents.detector.mrcnn import load_pretrained_model
from alfworld.agents.controller import OracleAgent, OracleAStarAgent, MaskRCNNAgent, MaskRCNNAStarAgent


TASK_TYPES = {1: "pick_and_place_simple",
              2: "look_at_obj_in_light",
              3: "pick_clean_then_place_in_recep",
              4: "pick_heat_then_place_in_recep",
              5: "pick_cool_then_place_in_recep",
              6: "pick_two_obj_and_place"}


class AlfredThorEnv(object):
    '''
    Interface for Embodied (THOR) environment
    '''

    class Thor(threading.Thread):
        def __init__(self, queue, train_eval="train"):
            Thread.__init__(self)
            self.action_queue = queue
            self.mask_rcnn = None
            self.env =  None
            self.train_eval = train_eval
            self.controller_type = "oracle"

        def run(self):
            while True:
                action, reset, task_file = self.action_queue.get()
                try:
                    if reset:
                        self.reset(task_file)
                    else:
                        self.step(action)
                finally:
                    self.action_queue.task_done()

        def init_env(self, config):
            self.config = config

            screen_height = config['env']['thor']['screen_height']
            screen_width = config['env']['thor']['screen_width']
            smooth_nav = config['env']['thor']['smooth_nav']
            save_frames_to_disk = config['env']['thor']['save_frames_to_disk']

            if not self.env:
                self.env = ThorEnv(player_screen_height=screen_height,
                                   player_screen_width=screen_width,
                                   smooth_nav=smooth_nav,
                                   save_frames_to_disk=save_frames_to_disk)
            self.controller_type = self.config['controller']['type']
            self._done = False
            self._res = ()
            self._feedback = ""
            self.expert = HandCodedThorAgent(self.env, max_steps=200)
            self.prev_command = ""
            self.load_mask_rcnn()

        def load_mask_rcnn(self):
            # load pretrained MaskRCNN model if required
            if 'mrcnn' in self.config['controller']['type'] and not self.mask_rcnn:
                model_path = os.path.expandvars(self.config['mask_rcnn']['pretrained_model_path'])
                self.mask_rcnn = load_pretrained_model(model_path)

        def set_task(self, task_file):
            self.task_file = task_file
            self.traj_root = os.path.dirname(task_file)
            with open(task_file, 'r') as f:
                self.traj_data = json.load(f)

        def reset(self, task_file):
            assert self.env
            assert self.controller_type

            self.set_task(task_file)

            # scene setup
            scene_num = self.traj_data['scene']['scene_num']
            object_poses = self.traj_data['scene']['object_poses']
            dirty_and_empty = self.traj_data['scene']['dirty_and_empty']
            object_toggles = self.traj_data['scene']['object_toggles']
            scene_name = 'FloorPlan%d' % scene_num
            self.env.reset(scene_name)
            self.env.restore_scene(object_poses, object_toggles, dirty_and_empty)

            # recording
            save_frames_path = self.config['env']['thor']['save_frames_path']
            self.env.save_frames_path = os.path.join(save_frames_path, self.traj_root.replace('../', ''))

            # initialize to start position
            self.env.step(dict(self.traj_data['scene']['init_action']))            # print goal instr
            task_desc = get_templated_task_desc(self.traj_data)
            print("Task: %s" % task_desc)
            # print("Task: %s" % (self.traj_data['turk_annotations']['anns'][0]['task_desc']))

            # setup task for reward
            class args: pass
            args.reward_config = os.path.join(alfworld.agents.__path__[0], 'config/rewards.json')
            self.env.set_task(self.traj_data, args, reward_type='dense')

            # set controller
            self.controller_type = self.config['controller']['type']
            self.goal_desc_human_anns_prob = self.config['env']['goal_desc_human_anns_prob']
            load_receps = self.config['controller']['load_receps']
            debug = self.config['controller']['debug']

            if self.controller_type == 'oracle':
                self.controller = OracleAgent(self.env, self.traj_data, self.traj_root,
                                              load_receps=load_receps, debug=debug,
                                              goal_desc_human_anns_prob=self.goal_desc_human_anns_prob)
            elif self.controller_type == 'oracle_astar':
                self.controller = OracleAStarAgent(self.env, self.traj_data, self.traj_root,
                                                   load_receps=load_receps, debug=debug,
                                                   goal_desc_human_anns_prob=self.goal_desc_human_anns_prob)
            elif self.controller_type == 'mrcnn':
                self.controller = MaskRCNNAgent(self.env, self.traj_data, self.traj_root,
                                                pretrained_model=self.mask_rcnn,
                                                load_receps=load_receps, debug=debug,
                                                goal_desc_human_anns_prob=self.goal_desc_human_anns_prob,
                                                save_detections_to_disk=self.env.save_frames_to_disk, save_detections_path=self.env.save_frames_path)
            elif self.controller_type == 'mrcnn_astar':
                self.controller = MaskRCNNAStarAgent(self.env, self.traj_data, self.traj_root,
                                                     pretrained_model=self.mask_rcnn,
                                                     load_receps=load_receps, debug=debug,
                                                     goal_desc_human_anns_prob=self.goal_desc_human_anns_prob,
                                                     save_detections_to_disk=self.env.save_frames_to_disk, save_detections_path=self.env.save_frames_path)
            else:
                raise NotImplementedError()

            # zero steps
            self.steps = 0

            # reset expert state
            self.expert.reset(task_file)
            self.prev_command = ""

            # return intro text
            self._feedback = self.controller.feedback
            self._res = self.get_info()

            return self._feedback

        def step(self, action):
            if not self._done:
                # take action
                self.prev_command = str(action)
                self._feedback = self.controller.step(action)
                self._res = self.get_info()
                if self.env.save_frames_to_disk:
                    self.record_action(action)
            self.steps += 1

        def get_results(self):
            return self._res

        def record_action(self, action):
            txt_file = os.path.join(self.env.save_frames_path, 'action.txt')
            with open(txt_file, 'a+') as f:
                f.write("%s\r\n" % str(action))

        def get_info(self):
            won = self.env.get_goal_satisfied()
            pcs = self.env.get_goal_conditions_met()
            goal_condition_success_rate = pcs[0] / float(pcs[1])
            acs = self.controller.get_admissible_commands()

            # expert action
            if self.train_eval == "train":
                game_state = {
                    'admissible_commands': acs,
                    'feedback': self._feedback,
                    'won': won
                }
                expert_actions = ["look"]
                try:
                    if not self.prev_command:
                        self.expert.observe(game_state['feedback'])
                    else:
                        next_action = self.expert.act(game_state, 0, won, self.prev_command)
                        if next_action in acs:
                            expert_actions = [next_action]
                except HandCodedAgentTimeout:
                    print("Expert Timeout")
                except Exception as e:
                    print(e)
                    traceback.print_exc()
            else:
                expert_actions = []

            training_method = self.config["general"]["training_method"]
            if training_method == "dqn":
                max_nb_steps_per_episode = self.config["rl"]["training"]["max_nb_steps_per_episode"]
            elif training_method == "dagger":
                max_nb_steps_per_episode = self.config["dagger"]["training"]["max_nb_steps_per_episode"]
            else:
                raise NotImplementedError
            self._done = won or self.steps > max_nb_steps_per_episode
            return (self._feedback, self._done, acs, won, goal_condition_success_rate, expert_actions)

        def get_last_frame(self):
            return self.env.last_event.frame[:,:,::-1]

        def get_exploration_frames(self):
            return self.controller.get_exploration_frames()

    def __init__(self, config, train_eval="train"):
        print("Initialize AlfredThorEnv...")
        self.config = config
        self.train_eval = train_eval
        self.random_seed = 123
        self.batch_size = 1
        self.envs = []
        self.action_queues = []
        self.get_env_paths()

    def close(self):
        for env in self.envs:
            env.env.stop()

    def seed(self, seed):
        self.random_seed = seed

    def get_env_paths(self):
        self.json_file_list = []

        if self.train_eval == "train":
            data_path = os.path.expandvars(self.config['dataset']['data_path'])
        elif self.train_eval == "eval_in_distribution":
            data_path = os.path.expandvars(self.config['dataset']['eval_id_data_path'])
        elif self.train_eval == "eval_out_of_distribution":
            data_path = os.path.expandvars(self.config['dataset']['eval_ood_data_path'])
        else:
            raise Exception("Invalid split. Must be either train or eval")

        # get task types
        assert len(self.config['env']['task_types']) > 0
        task_types = []
        for tt_id in self.config['env']['task_types']:
            if tt_id in TASK_TYPES:
                task_types.append(TASK_TYPES[tt_id])

        for root, dirs, files in os.walk(data_path, topdown=False):
            if 'traj_data.json' in files:
                # Skip movable and slice objects object tasks
                if 'movable' in root or 'Sliced' in root:
                    continue

                # File paths
                json_path = os.path.join(root, 'traj_data.json')
                game_file_path = os.path.join(root, "game.tw-pddl")

                # Load trajectory file
                with open(json_path, 'r') as f:
                    traj_data = json.load(f)

                # Check for any task_type constraints
                if not traj_data['task_type'] in task_types:
                    continue

                self.json_file_list.append(json_path)

                # # Only add solvable games
                # if os.path.exists(game_file_path):
                #     with open(game_file_path, 'r') as f:
                #         gamedata = json.load(f)
                #
                #     if 'solvable' in gamedata and gamedata['solvable']:
                #         self.json_file_list.append(json_path)

        print("Overall we have %s games..." % (str(len(self.json_file_list))))
        self.num_games = len(self.json_file_list)

        if self.train_eval == "train":
            num_train_games = self.config['dataset']['num_train_games'] if self.config['dataset']['num_train_games'] > 0 else len(self.json_file_list)
            self.json_file_list = self.json_file_list[:num_train_games]
            self.num_games = len(self.json_file_list)
            print("Training with %d games" % (len(self.json_file_list)))
        else:
            num_eval_games = self.config['dataset']['num_eval_games'] if self.config['dataset']['num_eval_games'] > 0 else len(self.json_file_list)
            self.json_file_list = self.json_file_list[:num_eval_games]
            self.num_games = len(self.json_file_list)
            print("Evaluating with %d games" % (len(self.json_file_list)))

    def init_env(self, batch_size):
        self.get_env_paths()
        self.batch_size = batch_size
        self.action_queues = []
        self.task_order = [""] * self.batch_size
        for n in range(self.batch_size):
            queue = Queue()
            env = self.Thor(queue, self.train_eval)
            self.action_queues.append(queue)
            self.envs.append(env)
            env.daemon = True
            env.start()
            env.init_env(self.config)
        return self

    def reset(self):
        # set tasks
        batch_size = self.batch_size
        # reset envs

        if self.train_eval == 'train':
            tasks = random.sample(self.json_file_list, k=batch_size)
        else:
            if len(self.json_file_list)-batch_size > batch_size:
                tasks = [self.json_file_list.pop(random.randrange(len(self.json_file_list))) for _ in range(batch_size)]
            else:
                tasks = random.sample(self.json_file_list, k=batch_size)
                self.get_env_paths()

        for n in range(batch_size):
            self.action_queues[n].put((None, True, tasks[n]))

        obs, dones, infos = self.wait_and_get_info()
        return obs, infos

    def step(self, actions):
        '''
        executes actions in parallel and waits for all env to finish
        '''

        batch_size = self.batch_size
        for n in range(batch_size):
            self.action_queues[n].put((actions[n], False, ""))

        obs, dones, infos = self.wait_and_get_info()
        return obs, None, dones, infos

    def wait_and_get_info(self):
        obs, dones, admissible_commands, wons, gamefiles, expert_plans, gc_srs = [], [], [], [], [], [], []

        # wait for all threads
        for n in range(self.batch_size):
            self.action_queues[n].join()
            feedback, done, acs, won, gc_sr, expert_actions = self.envs[n].get_results()
            obs.append(feedback)
            dones.append(done)
            admissible_commands.append(acs)
            wons.append(won)
            gc_srs.append(gc_sr)
            gamefiles.append(self.envs[n].traj_root)
            expert_plans.append(expert_actions)

        infos = {'admissible_commands': admissible_commands,
                 'won': wons,
                 'goal_condition_success_rate': gc_srs,
                 'extra.gamefile': gamefiles,
                 'expert_plan': expert_plans}
        return obs, dones, infos

    def get_frames(self):
        images = []
        for n in range(self.batch_size):
            images.append(self.envs[n].get_last_frame())
        return np.array(images)

    def get_exploration_frames(self):
        images = []
        for n in range(self.batch_size):
            images.append(self.envs[n].get_exploration_frames())
        return images


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.


import os

from setuptools import setup, find_packages

setup(
    name='alfworld',
    version=open(os.path.join("alfworld", "info.py")).readlines()[0].split("=")[-1].strip("' \n"),
    packages=find_packages(),
    scripts=[
        "scripts/alfworld-download",
        "scripts/alfworld-play-tw",
        "scripts/alfworld-play-thor",
    ],
    include_package_data=True,
    license=open('LICENSE').read(),
    zip_safe=False,
    description="ALFWorld - Aligning Text and Embodied Environments for Interactive Learning.",
    install_requires=[line for line in open('requirements.txt').readlines() if "@" not in line],
)



import os
import sys
import copy

import numpy as np
import torch
import torch.nn.functional as F

import alfworld.agents
import alfworld.agents.modules.memory as memory
from alfworld.agents.agent import TextDAggerAgent
from alfworld.agents.modules.generic import to_np, to_pt, _words_to_ids, pad_sequences, preproc, max_len, ez_gather_dim_1, LinearSchedule
from alfworld.agents.modules.layers import NegativeLogLoss, masked_mean, compute_mask
from alfworld.agents.detector.mrcnn import load_pretrained_model

import torchvision.transforms as T
from torchvision import models
from torchvision.ops import boxes as box_ops


class VisionDAggerAgent(TextDAggerAgent):
    '''
    Vision Agent trained with DAgger
    '''
    def __init__(self, config):
        super().__init__(config)

        assert self.action_space == "generation"

        self.use_gpu = config['general']['use_cuda']
        self.transform = T.Compose([T.ToTensor()])

        # choose vision model
        self.vision_model_type = config['vision_dagger']['model_type']
        self.use_exploration_frame_feats = config['vision_dagger']['use_exploration_frame_feats']
        self.sequence_aggregation_method = config['vision_dagger']['sequence_aggregation_method']

        # initialize model
        if self.vision_model_type in {'resnet'}:
            self.detector = models.resnet18(pretrained=True)
            self.detector.eval()
            if self.use_gpu:
                self.detector.cuda()
        elif self.vision_model_type in {'maskrcnn', 'maskrcnn_whole'}:
            pretrained_model_path = config['mask_rcnn']['pretrained_model_path']
            self.mask_rcnn_top_k_boxes = self.config['vision_dagger']['maskrcnn_top_k_boxes']
            self.avg2dpool = torch.nn.AvgPool2d((13, 13))
            self.detector = load_pretrained_model(pretrained_model_path)
            self.detector.roi_heads.register_forward_hook(self.box_features_hook)
            self.detection_box_features = []
            self.fpn_pooled_features = []
            self.detector.eval()
            if self.use_gpu:
                self.detector.cuda()
        elif self.vision_model_type in {"no_vision"}:
            print("No Vision Agent")
        else:
            raise NotImplementedError()

    def box_features_hook(self, module, input, output):
        '''
        hook for extracting features from MaskRCNN
        '''

        features, proposals, image_shapes, targets = input

        box_features = module.box_roi_pool(features, proposals, image_shapes)
        box_features = module.box_head(box_features)
        class_logits, box_regression = module.box_predictor(box_features)

        device = class_logits.device
        num_classes = class_logits.shape[-1]

        boxes_per_image = [len(boxes_in_image) for boxes_in_image in proposals]
        pred_boxes = module.box_coder.decode(box_regression, proposals)

        pred_scores = F.softmax(class_logits, -1)

        # split boxes and scores per image
        pred_boxes = pred_boxes.split(boxes_per_image, 0)
        pred_scores = pred_scores.split(boxes_per_image, 0)

        all_boxes = []
        all_scores = []
        all_labels = []
        all_keeps = []
        for boxes, scores, image_shape in zip(pred_boxes, pred_scores, image_shapes):
            boxes = box_ops.clip_boxes_to_image(boxes, image_shape)

            # create labels for each prediction
            labels = torch.arange(num_classes, device=device)
            labels = labels.view(1, -1).expand_as(scores)

            # remove predictions with the background label
            boxes = boxes[:, 1:]
            scores = scores[:, 1:]
            labels = labels[:, 1:]

            # batch everything, by making every class prediction be a separate instance
            boxes = boxes.reshape(-1, 4)
            scores = scores.flatten()
            labels = labels.flatten()

            # remove low scoring boxes
            inds = torch.nonzero(scores > module.score_thresh).squeeze(1)
            boxes, scores, labels = boxes[inds], scores[inds], labels[inds]

            # remove empty boxes
            keep = box_ops.remove_small_boxes(boxes, min_size=1e-2)
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            # non-maximum suppression, independently done per class
            keep = box_ops.batched_nms(boxes, scores, labels, module.nms_thresh)
            # keep only topk scoring predictions
            keep = keep[:self.mask_rcnn_top_k_boxes]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            all_boxes.append(boxes)
            all_scores.append(scores)
            all_labels.append(labels)
            all_keeps.append(keep)

        box_features_per_image = []
        for keep in all_keeps:
            box_features_per_image.append(box_features[keep])

        self.detection_box_features = box_features_per_image
        self.fpn_pooled_features = self.avg2dpool(features['pool']).squeeze(-1).squeeze(-1)

    # visual features for state representation
    def extract_visual_features(self, images):
        with torch.no_grad():
            if "resnet" in self.vision_model_type:
                image_tensors = [self.transform(i).cuda() if self.use_gpu else self.transform() for i in images]
                image_tensors = torch.stack(image_tensors, dim=0)
                res_out = self.detector(image_tensors)
                res_out_list = [res_out[i].unsqueeze(0) for i in range(res_out.shape[0])]
                return res_out_list
            elif "maskrcnn" in self.vision_model_type:
                image_tensors = [self.transform(i).cuda() if self.use_gpu else self.transform(i) for i in images]
                self.detector(image_tensors) # hook writes to self.detection_box_features
                if "maskrcnn_whole" in self.vision_model_type:
                    return [i.unsqueeze(0) for i in self.fpn_pooled_features]
                else:
                    return self.detection_box_features
            elif "no_vision" in self.vision_model_type:
                batch_size = len(images)
                zeros = [torch.zeros((1, 1000)) for _ in range(batch_size)]
                if self.use_gpu:
                    zeros = [z.cuda() for z in zeros]
                return zeros
            else:
                raise NotImplementedError()

    # without recurrency
    def train_dagger(self):

        if len(self.dagger_memory) < self.dagger_replay_batch_size:
            return None
        transitions = self.dagger_memory.sample(self.dagger_replay_batch_size)
        if transitions is None:
            return None
        batch = memory.dagger_transition(*zip(*transitions))

        if self.action_space == "generation":
            return self.command_generation_teacher_force(batch.observation_list, batch.task_list, batch.target_list)
        else:
            raise NotImplementedError()

    # with recurrency
    def train_dagger_recurrent(self):

        if len(self.dagger_memory) < self.dagger_replay_batch_size:
            return None
        sequence_of_transitions, contains_first_step = self.dagger_memory.sample_sequence(self.dagger_replay_batch_size, self.dagger_replay_sample_history_length)
        if sequence_of_transitions is None:
            return None

        batches = []
        for transitions in sequence_of_transitions:
            batch = memory.dagger_transition(*zip(*transitions))
            batches.append(batch)

        if self.action_space == "generation":
            return self.command_generation_recurrent_teacher_force([batch.observation_list for batch in batches], [batch.task_list for batch in batches], [batch.target_list for batch in batches], contains_first_step)
        else:
            raise NotImplementedError()

    def command_generation_teacher_force(self, observation_feats, task_desc_strings, target_strings):
        input_target_strings = [" ".join(["[CLS]"] + item.split()) for item in target_strings]
        output_target_strings = [" ".join(item.split() + ["[SEP]"]) for item in target_strings]
        batch_size = len(observation_feats)

        aggregated_obs_feat = self.aggregate_feats_seq(observation_feats)
        h_obs = self.online_net.vision_fc(aggregated_obs_feat)
        h_td, td_mask = self.encode(task_desc_strings, use_model="online")
        h_td_mean = self.online_net.masked_mean(h_td, td_mask).unsqueeze(1)
        h_obs = h_obs.to(h_td_mean.device)
        vision_td = torch.cat((h_obs, h_td_mean), dim=1) # batch x k boxes x hi
        vision_td_mask = torch.ones((batch_size, h_obs.shape[1]+h_td_mean.shape[1])).to(h_td_mean.device)

        input_target = self.get_word_input(input_target_strings)
        ground_truth = self.get_word_input(output_target_strings)  # batch x target_length
        target_mask = compute_mask(input_target)  # mask of ground truth should be the same
        pred = self.online_net.vision_decode(input_target, target_mask, vision_td, vision_td_mask, None)  # batch x target_length x vocab

        batch_loss = NegativeLogLoss(pred * target_mask.unsqueeze(-1), ground_truth, target_mask, smoothing_eps=self.smoothing_eps)
        loss = torch.mean(batch_loss)

        if loss is None:
            return None, None
        # Backpropagate
        self.online_net.zero_grad()
        self.optimizer.zero_grad()
        loss.backward()
        # `clip_grad_norm` helps prevent the exploding gradient problem in RNNs / LSTMs.
        torch.nn.utils.clip_grad_norm_(self.online_net.parameters(), self.clip_grad_norm)
        self.optimizer.step()  # apply gradients
        return to_np(pred), to_np(loss)

    def command_generation_recurrent_teacher_force(self, seq_observation_feats, seq_task_desc_strings, seq_target_strings, contains_first_step=False):
        loss_list = []
        previous_dynamics = None
        batch_size = len(seq_observation_feats[0])
        h_td, td_mask = self.encode(seq_task_desc_strings[0], use_model="online")
        h_td_mean = self.online_net.masked_mean(h_td, td_mask).unsqueeze(1)
        for step_no in range(self.dagger_replay_sample_history_length):
            input_target_strings = [" ".join(["[CLS]"] + item.split()) for item in seq_target_strings[step_no]]
            output_target_strings = [" ".join(item.split() + ["[SEP]"]) for item in seq_target_strings[step_no]]

            obs = [o.to(h_td.device) for o in seq_observation_feats[step_no]]
            aggregated_obs_feat = self.aggregate_feats_seq(obs)
            h_obs = self.online_net.vision_fc(aggregated_obs_feat)
            vision_td = torch.cat((h_obs, h_td_mean), dim=1) # batch x k boxes x hid
            vision_td_mask = torch.ones((batch_size, h_obs.shape[1]+h_td_mean.shape[1])).to(h_td_mean.device)

            averaged_vision_td_representation = self.online_net.masked_mean(vision_td, vision_td_mask)
            current_dynamics = self.online_net.rnncell(averaged_vision_td_representation, previous_dynamics) if previous_dynamics is not None else self.online_net.rnncell(averaged_vision_td_representation)

            input_target = self.get_word_input(input_target_strings)
            ground_truth = self.get_word_input(output_target_strings)  # batch x target_length
            target_mask = compute_mask(input_target)  # mask of ground truth should be the same
            pred = self.online_net.vision_decode(input_target, target_mask, vision_td, vision_td_mask, current_dynamics)  # batch x target_length x vocab

            previous_dynamics = current_dynamics
            if (not contains_first_step) and step_no < self.dagger_replay_sample_update_from:
                previous_dynamics = previous_dynamics.detach()
                continue

            batch_loss = NegativeLogLoss(pred * target_mask.unsqueeze(-1), ground_truth, target_mask, smoothing_eps=self.smoothing_eps)
            loss = torch.mean(batch_loss)
            loss_list.append(loss)

        loss = torch.stack(loss_list).mean()
        if loss is None:
            return None
        # Backpropagate
        self.online_net.zero_grad()
        self.optimizer.zero_grad()
        loss.backward()
        # `clip_grad_norm` helps prevent the exploding gradient problem in RNNs / LSTMs.
        torch.nn.utils.clip_grad_norm_(self.online_net.parameters(), self.clip_grad_norm)
        self.optimizer.step()  # apply gradients
        return to_np(loss)

    def command_generation_greedy_generation(self, observation_feats, task_desc_strings, previous_dynamics):
        with torch.no_grad():
            batch_size = len(observation_feats)

            aggregated_obs_feat = self.aggregate_feats_seq(observation_feats)
            h_obs = self.online_net.vision_fc(aggregated_obs_feat)
            h_td, td_mask = self.encode(task_desc_strings, use_model="online")
            h_td_mean = self.online_net.masked_mean(h_td, td_mask).unsqueeze(1)
            h_obs = h_obs.to(h_td_mean.device)
            vision_td = torch.cat((h_obs, h_td_mean), dim=1) # batch x k boxes x hid
            vision_td_mask = torch.ones((batch_size, h_obs.shape[1]+h_td_mean.shape[1])).to(h_td_mean.device)

            if self.recurrent:
                averaged_vision_td_representation = self.online_net.masked_mean(vision_td, vision_td_mask)
                current_dynamics = self.online_net.rnncell(averaged_vision_td_representation, previous_dynamics) if previous_dynamics is not None else self.online_net.rnncell(averaged_vision_td_representation)
            else:
                current_dynamics = None

            # greedy generation
            input_target_list = [[self.word2id["[CLS]"]] for i in range(batch_size)]
            eos = np.zeros(batch_size)
            for _ in range(self.max_target_length):

                input_target = copy.deepcopy(input_target_list)
                input_target = pad_sequences(input_target, maxlen=max_len(input_target)).astype('int32')
                input_target = to_pt(input_target, self.use_cuda)
                target_mask = compute_mask(input_target)  # mask of ground truth should be the same
                pred = self.online_net.vision_decode(input_target, target_mask, vision_td, vision_td_mask, current_dynamics)  # batch x target_length x vocab
                # pointer softmax
                pred = to_np(pred[:, -1])  # batch x vocab
                pred = np.argmax(pred, -1)  # batch
                for b in range(batch_size):
                    new_stuff = [pred[b]] if eos[b] == 0 else []
                    input_target_list[b] = input_target_list[b] + new_stuff
                    if pred[b] == self.word2id["[SEP]"]:
                        eos[b] = 1
                if np.sum(eos) == batch_size:
                    break
            res = [self.tokenizer.decode(item) for item in input_target_list]
            res = [item.replace("[CLS]", "").replace("[SEP]", "").strip() for item in res]
            res = [item.replace(" in / on ", " in/on " ) for item in res]
            return res, current_dynamics

    def get_vision_feat_mask(self, observation_feats):
        batch_size = len(observation_feats)
        num_vision_feats = [of.shape[0] for of in observation_feats]
        max_feat_len = max(num_vision_feats)
        mask = torch.zeros((batch_size, max_feat_len))
        for b, num_vision_feat in enumerate(num_vision_feats):
            mask[b,:num_vision_feat] = 1
        return mask

    def extract_exploration_frame_feats(self, exploration_frames):
        exploration_frame_feats = []
        for batch in exploration_frames:
            ef_feats = []
            for image in batch:
                ef_feats.append(self.extract_visual_features([image])[0])
            # cat_feats = torch.cat(ef_feats, dim=0)
            max_feat_len = max([f.shape[0] for f in ef_feats])
            stacked_feats = self.online_net.vision_fc.pad_and_stack(ef_feats, max_feat_len=max_feat_len)
            stacked_feats = stacked_feats.view(-1, self.online_net.vision_fc.in_features)
            exploration_frame_feats.append(stacked_feats)
        return exploration_frame_feats

    def aggregate_feats_seq(self, feats):
        if self.sequence_aggregation_method == "sum":
            return [f.sum(0).unsqueeze(0) for f in feats]
        elif self.sequence_aggregation_method == "average":
            return [f.mean(0).unsqueeze(0) for f in feats]
        elif self.sequence_aggregation_method == "rnn":
            max_feat_len = max([f.shape[0] for f in feats])
            feats_stack = self.online_net.vision_fc.pad_and_stack(feats, max_feat_len=max_feat_len)
            feats_h, feats_c = self.online_net.vision_feat_seq_rnn(feats_stack)
            aggregated_feats = feats_h[:,0,:].unsqueeze(1)
            return [b for b in aggregated_feats]
        else:
            raise ValueError("sequence_aggregation_method must be sum, average or rnn")



import hashlib
import random
import string

from collections import Counter
from textworld.logic import Proposition, Variable

import os
import sys
import alfworld.gen.goal_library as glib
import alfworld.gen.constants as constants

import numpy as np
from json import JSONEncoder


class Demangler(object):

    def __init__(self, obj_dict=None, game_infos=None, shuffle=False):
        if obj_dict is None:
            self.obj_count = Counter()
        else:
            self.obj_count = obj_dict

        self.obj_names = {}
        if game_infos:
            ids = sorted([info.id for info in game_infos.values()])
            if shuffle:
                random.shuffle(ids)

            # count the number of instances
            for id in ids:
                splits = id.split("_bar_", 1)
                if len(splits) > 1:
                    name, rest = splits
                    if "basin" in id:
                        name += "basin"
                    self.obj_count[name] += 1

            # make list of num ids for each object (shuffle the ids if required)
            obj_num_ids = {}
            for obj, count in self.obj_count.most_common():
                num_ids = list(range(count+1)[1:])  # start from index 1
                obj_num_ids[obj] = num_ids

            # assign unique num id for each object based on precomputed list
            for id in ids:
                text = id
                text = text.replace("_bar_", "|")
                text = text.replace("_minus_", "-")
                text = text.replace("_plus_", "+")
                text = text.replace("_dot_", ".")
                text = text.replace("_comma_", ",")

                splits = text.split("|", 1)
                if len(splits) == 1:
                    self.obj_names[id] = {'name': text, 'id': 0}
                else:
                    name, rest = splits
                    if "basin" in id:
                        name += "basin"
                    self.obj_names[id] = {'name': name, 'id': obj_num_ids[name].pop()}

    def demangle_alfred_name(self, text):
        assert(text in self.obj_names)
        name, id = self.obj_names[text].values()
        id = str(id) if id > 0 else ""
        res = "{} {}".format(name, id)
        return res


class NumpyArrayEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return JSONEncoder.default(self, obj)


def get_templated_task_desc(traj_data):

    pddl_params = traj_data['pddl_params']
    goal_str = traj_data['task_type']
    if pddl_params['object_sliced']:
        goal_str += "_slice"

    template = random.choice(glib.gdict[goal_str]['templates'])
    obj = pddl_params['object_target'].lower()
    recep = pddl_params['parent_target'].lower()
    toggle = pddl_params['toggle_target'].lower()
    mrecep = pddl_params['mrecep_target'].lower()
    filled_in_str = template.format(obj=obj, recep=recep, toggle=toggle, mrecep=mrecep)
    return filled_in_str


def clean_alfred_facts(facts):
    # TODO: fix this?
    demangler = Demangler()

    def _clean_fact(fact: Proposition):
        args = [Variable(demangler.demangle_alfred_name(arg.name), arg.type) for arg in fact.arguments]
        return Proposition(fact.name, args)

    facts = [_clean_fact(fact) for fact in facts if not fact.name.startswith("new-axiom@")]
    return facts


def add_task_to_grammar(grammar_str, traj_data, goal_desc_human_anns_prob=0.0):
    if random.random() < goal_desc_human_anns_prob:
        task_desc = get_human_anns_task_desc(traj_data)
    else:
        task_desc = get_templated_task_desc(traj_data)
    return grammar_str.replace("UNKNOWN GOAL", task_desc)


def get_human_anns_task_desc(traj_data):
    return random.choice(traj_data['turk_annotations']['anns'])['task_desc']


def parse_objects(desc):
    '''
    extract objects after "you see" and before "your task is to:"
    '''
    # TODO: replace with a noun-phrase extractor
    objs = []
    if "you see" in desc.lower():
        obj_str = desc.lower().replace(" and", ",")
        obj_str = obj_str.split("you see ", 1)[1]
        obj_str = obj_str.partition("your task is to:")[0]
        for s in obj_str.split(","):
            item = s.translate(str.maketrans('', '', string.punctuation))
            item = item.strip().replace("a ", "")
            if item:
                objs.append(item)
    return objs


def extract_admissible_commands_with_heuristics(intro, frame_desc, feedback,
                                                curr_recep, inventory):
    '''
    Heavily engineered admissible commands extraction. Lots of commonsense and heuristics used to prune list.
    '''

    at_recep = str(curr_recep) if curr_recep != "nothing" else ""
    at_recep_type = at_recep.split()[0] if at_recep else ""
    in_inv = str(inventory[0]).lower() if len(inventory) > 0 else ""

    OPENABLE_RECEPTACLES = [r.lower() for r in constants.OPENABLE_CLASS_LIST]

    assert intro  # intro should be non-empty

    admissible_commands = []
    templates = [
        "go to {recep}",
        "open {recep}",
        "close {recep}",
        "take {obj} from {recep}",
        "put {obj} in/on {recep}",
        "use {lamp}",
        "heat {obj} with {microwave}",
        "cool {obj} with {fridge}",
        "clean {obj} with {cleaner}",
        "slice {obj} with {knife}",
        "inventory",
        "look",
        "examine {obj}",
        "examine {recep}"
    ]

    # parse interactable and navigable objects
    receps = parse_objects(intro)
    objects = parse_objects(frame_desc) if frame_desc else []
    lamps = [obj for obj in objects if 'lamp' in obj]
    microwaves = [obj for obj in receps if 'microwave' in obj]
    cleaners = [obj for obj in receps if 'sink' in obj or 'bathtub' in obj]
    fridges = [obj for obj in receps if 'fridge' in obj]
    knives = [obj for obj in receps if 'knife' in obj]

    # populate templates
    for t in templates:
        if 'take {obj} from {recep}' in t:
            if at_recep and not in_inv:
                for obj in objects:
                    if 'desklamp' not in obj and 'floorlamp' not in obj:
                        admissible_commands.append(t.format(recep=at_recep, obj=obj))
        elif 'put {obj} in/on {recep}' in t:
            if in_inv and at_recep:
                admissible_commands.append(t.format(recep=at_recep, obj=in_inv))
        elif '{obj}' in t and '{microwave}' in t:
            if 'microwave' in at_recep and in_inv:
                for microwave in microwaves:
                    admissible_commands.append(t.format(microwave=microwave, obj=in_inv))
        elif '{obj}' in t and '{fridge}' in t:
            if 'fridge' in at_recep and in_inv:
                for fridge in fridges:
                    admissible_commands.append(t.format(fridge=fridge, obj=in_inv))
        elif '{obj}' in t and '{cleaner}' in t:
            if ('sink' in at_recep or 'bathtub' in at_recep) and in_inv:
                for cleaner in cleaners:
                    admissible_commands.append(t.format(cleaner=cleaner, obj=in_inv))
        elif '{obj}' in t and '{knife}' in t:
            if 'knife' in in_inv:
                for knife in knives:
                    for obj in objects:
                        admissible_commands.append(t.format(knife=knife, obj=obj))
        elif 'open {recep}' in t:
            if at_recep:
                if at_recep_type in OPENABLE_RECEPTACLES: # and ("is closed" in feedback or "You close" in feedback):
                    admissible_commands.append(t.format(recep=at_recep))
        elif 'close {recep}' in t:
            if at_recep:
                if at_recep_type in OPENABLE_RECEPTACLES: # and "is open" in feedback:
                    admissible_commands.append(t.format(recep=at_recep))
        elif 'examine {recep}' in t:
            if at_recep:
                admissible_commands.append(t.format(recep=at_recep))
        elif 'examine {obj}' in t:
            if in_inv:
                admissible_commands.append(t.format(obj=in_inv))
        elif 'go to {recep}' in t:
            for recep in receps:
                if recep != at_recep:
                    admissible_commands.append(t.format(recep=recep))
        elif '{recep}' in t:
            for recep in receps:
                admissible_commands.append(t.format(recep=recep))
        elif '{obj}' in t:
            for obj in objects:
                admissible_commands.append(t.format(obj=obj))
        elif '{lamp}' in t:
            for lamp in lamps:
                admissible_commands.append(t.format(lamp=lamp))
        else:
            admissible_commands.append(t)

    return admissible_commands


def extract_admissible_commands(intro, frame_desc):
    '''
    exhaustive list of admissible commands
    '''

    admissible_commands = []
    templates = [
        "go to {recep}",
        "open {recep}",
        "close {recep}",
        "take {obj} from {recep}",
        "put {obj} in/on {recep}",
        "use {lamp}",
        "heat {obj} with {microwave}",
        "cool {obj} with {fridge}",
        "clean {obj} with {cleaner}",
        "slice {obj} with {knife}",
        "inventory",
        "look",
        "examine {obj}",
        "examine {recep}"
    ]

    # parse interactable and navigable objects
    receps = parse_objects(intro)
    objects = parse_objects(frame_desc) if frame_desc else []
    lamps = [obj for obj in objects if 'lamp' in obj]
    microwaves = [obj for obj in receps if 'microwave' in obj]
    cleaners = [obj for obj in receps if 'sink' in obj or 'bathtub' in obj]
    fridges = [obj for obj in receps if 'fridge' in obj]
    knives = [obj for obj in receps if 'knife' in obj]

    # populate templates
    for t in templates:
        if '{recep}' in t and '{obj}' in t:
            for recep in receps:
                for obj in objects:
                    admissible_commands.append(t.format(recep=recep, obj=obj))
        elif '{obj}' in t and '{microwave}' in t:
            for microwave in microwaves:
                for obj in objects:
                    admissible_commands.append(t.format(microwave=microwave, obj=obj))
        elif '{obj}' in t and '{fridge}' in t:
            for fridge in fridges:
                for obj in objects:
                    admissible_commands.append(t.format(fridge=fridge, obj=obj))
        elif '{obj}' in t and '{cleaner}' in t:
            for cleaner in cleaners:
                for obj in objects:
                    admissible_commands.append(t.format(cleaner=cleaner, obj=obj))
        elif '{obj}' in t and '{knife}' in t:
            for knife in knives:
                for obj in objects:
                    admissible_commands.append(t.format(knife=knife, obj=obj))
        elif '{recep}' in t:
            for recep in receps:
                admissible_commands.append(t.format(recep=recep))
        elif '{obj}' in t:
            for obj in objects:
                admissible_commands.append(t.format(obj=obj))
        elif '{lamp}' in t:
            for lamp in lamps:
                admissible_commands.append(t.format(lamp=lamp))
        else:
            admissible_commands.append(t)

    return admissible_commands



import json
import os
import alfworld.gen
import alfworld.gen.constants as constants
from alfworld.gen.game_states.task_game_state_full_knowledge import TaskGameStateFullKnowledge
from alfworld.gen.agents.deterministic_planner_agent import DeterministicPlannerAgent
from alfworld.gen.graph import graph_obj
from alfworld.agents.controller.mrcnn import MaskRCNNAgent

class MaskRCNNAStarAgent(MaskRCNNAgent):

    def __init__(self, env, traj_data, traj_root,
                 pretrained_model=None,
                 load_receps=False, debug=False,
                 goal_desc_human_anns_prob=0.0,
                 classes=constants.OBJECTS_DETECTOR,
                 save_detections_to_disk=False, save_detections_path='./'):

        super().__init__(env, traj_data, traj_root,
                         pretrained_model=pretrained_model,
                         load_receps=load_receps, debug=debug,
                         goal_desc_human_anns_prob=goal_desc_human_anns_prob,
                         classes=classes,
                         save_detections_to_disk=save_detections_to_disk, save_detections_path=save_detections_path)

    def setup_navigator(self):
        game_state = TaskGameStateFullKnowledge(self.env)

        # reset
        game_state.receptacle_to_point = None
        game_state.task_target = None
        game_state.success = False

        # load nav graph
        scene_num = self.traj_data['scene']['scene_num']
        game_state.gt_graph = graph_obj.Graph(use_gt=True, construct_graph=True, scene_id=scene_num)
        game_state.gt_graph.clear()

        game_state.agent_height = self.env.last_event.metadata['agent']['position']['y']
        game_state.camera_height = game_state.agent_height + constants.CAMERA_HEIGHT_OFFSET

        points_source = os.path.join(alfworld.gen.__path__[0], 'layouts/FloorPlan%s-openable.json' % scene_num)
        with open(points_source, 'r') as f:
            openable_object_to_point = json.load(f)
        game_state.openable_object_to_point = openable_object_to_point

        game_state.update_receptacle_nearest_points() # TODO: save to desk
        game_state.planner.process_pool.terminate()

        self.navigator = DeterministicPlannerAgent(thread_id=0, game_state=game_state)


    def navigate(self, teleport_loc):
        self.navigator.pose = self.env.last_event.pose_discrete
        self.navigator.step(teleport_loc)
        return self.env.last_event


import os
import random

import alfworld.gen.constants as constants
import alfworld.gen.goal_library as glib
from alfworld.gen.game_states.planned_game_state import PlannedGameState
from alfworld.gen.utils import game_util


class TaskGameState(PlannedGameState):
    action_space = [
        {'action': 'Explore'},
        {'action': 'Scan'},
        {'action': 'Plan'},
        {'action': 'End'},
    ]

    def __init__(self, env, seed=None, action_space=None):
        if action_space is None:
            action_space = TaskGameState.action_space
        super(TaskGameState, self).__init__(env, seed, action_space,
                                            'put_task', 'planner/domains/PutTaskExtended_domain.pddl')
        self.task_target = None
        self.success = False

    def get_task_str(self):
        return game_util.get_task_str(self.object_target, self.parent_target, self.toggle_target, self.mrecep_target)

    def get_goal_pddl(self):
        goal_type = constants.pddl_goal_type
        if constants.data_dict['pddl_params']['object_sliced']:
            goal_type += "_slice"
        goal_str = glib.gdict[goal_type]['pddl']
        goal_str = goal_str.format(obj=constants.OBJECTS[self.object_target]
                                    if self.object_target is not None else "",
                                   recep=constants.OBJECTS[self.parent_target]
                                    if self.parent_target is not None else "",
                                   toggle=constants.OBJECTS[self.toggle_target]
                                    if self.toggle_target is not None else "",
                                   mrecep=constants.OBJECTS[self.mrecep_target]
                                    if self.mrecep_target is not None else "")
        return goal_str

    def get_current_plan(self, force_update=False):
        plan = super(TaskGameState, self).get_current_plan(force_update)
        self.plan = plan
        return self.plan

    def get_random_task_vals(self, scene=None):
        dataset_type, max_num_repeats, remove_prob, task_row = \
            self.initialize_random_scene(scene=scene)
        return (dataset_type, task_row), max_num_repeats, remove_prob

    def initialize_random_scene(self, scene=None):
        if self.problem_id is not None:
            # clean up old problem
            if not constants.DEBUG and os.path.exists('%s/planner/generated_problems/problem_%s.pddl' % (
                    self.dname, self.problem_id)):
                os.remove('%s/planner/generated_problems/problem_%s.pddl' % (self.dname, self.problem_id))

        # Do equal number of each task type in train.
        dataset_type = 'train'
        task_row = self.local_random.randint(0, 1e5)

        self.problem_id = str(task_row)

        if scene is None:
            # setup random scene with random object initialization
            self.scene_num = random.choice(constants.TRAIN_SCENE_NUMBERS)
            self.scene_seed = self.local_random.randint(0, 2 ** 32)
        else:
            # load a pre-specified sscene
            self.scene_num = scene['scene_num']
            self.scene_seed = scene['random_seed']

        self.object_target = -1
        self.parent_target = -1

        max_num_repeats = constants.MAX_NUM_OF_OBJ_INSTANCES
        remove_prob = 0.0
        return dataset_type, max_num_repeats, remove_prob, task_row

    def get_filter_crit(self, goal_type):

        # helper functions
        def is_obj_pickupable(x):
            return x['pickupable'] and x['objectType'] in constants.OBJECTS

        def is_receptacle(x):
            return x['receptacle'] and x['objectType'] in constants.OBJECTS

        def does_have_duplicates(x):
            num_instances = len(game_util.get_objects_of_type(x['objectType'], self.env.last_event.metadata))
            return num_instances > 1

        def does_have_atleast_n_duplicates(x, n):
            num_instances = len(game_util.get_objects_of_type(x['objectType'], self.env.last_event.metadata))
            return num_instances >= n

        def does_any_recep_type_have_obj_of_type(r, ot):
            all_recep_of_type = [o for o in self.env.last_event.metadata['objects'] if o['objectType'] == r['objectType']]
            valid_recep = []
            for recep in all_recep_of_type:
                receptacle_obj_ids = recep['receptacleObjectIds']
                if receptacle_obj_ids is not None:
                    objs_of_type = [obj_id for obj_id in receptacle_obj_ids if ot in obj_id]
                    if len(objs_of_type) > 0:
                        valid_recep.append(recep)
            return len(valid_recep) > 0

        def does_recep_have_enough_space(r):
            # TODO: how to check this with THOR 2.1.0?
            return True

        def is_recep_full(r):
            # TODO: how to check this with THOR 2.1.0?
            return False

        def can_recep_hold_obj(o, r):
            return o in constants.VAL_RECEPTACLE_OBJECTS[r['objectType']]

        def is_of_type(x, type):
            return x['objectType'] == type

        def is_movable_recep(o):
            return o['objectType'] in constants.MOVABLE_RECEPTACLES

        def is_obj_type_in_recep(x, recep):
            obj_types_in_recep = [o['objectType'] for o in self.env.last_event.metadata['objects']
                                  if (o['parentReceptacles'] is not None and
                                      any(recep in r for r in o['parentReceptacles']))]
            return x['objectType'] in obj_types_in_recep

        def is_obj_in_recep(o, r):
            return o['objectId'] in r['receptacleObjectIds']

        def is_obj_prop(x, prop):
            return x['objectType'] in constants.VAL_ACTION_OBJECTS[prop]

        def is_obj_off(x):
            return x['objectType'] in constants.VAL_ACTION_OBJECTS['Toggleable'] and \
                   x['toggleable'] \
                   and (not x['isToggled'])

        def is_obj_on(x):
            return x['objectType'] in constants.VAL_ACTION_OBJECTS['Toggleable'] and \
                   x['toggleable'] and \
                   x['isToggled']

        # specialized filters for each specific task
        if goal_type == "init":
            return lambda o: is_obj_pickupable(o), \
                   lambda r: is_receptacle(r) and \
                             (can_recep_hold_obj(self.rand_chosen_object_class, r)) and \
                             (not is_recep_full(r)) and \
                             (is_obj_in_recep(self.rand_chosen_object, r))
        elif goal_type == "place_all_obj_type_into_recep":
            return lambda o: does_have_duplicates(o), \
                   lambda r: does_recep_have_enough_space(r)
        elif goal_type == "pick_two_obj_and_place":
            return lambda o: does_have_atleast_n_duplicates(o, 2), \
                   lambda r: does_recep_have_enough_space(r) and \
                             (not does_any_recep_type_have_obj_of_type(r, self.rand_chosen_object_class))
        elif goal_type == "pick_clean_then_place_in_recep":
            return lambda o: (not is_obj_type_in_recep(o, "SinkBasin")) and \
                             (is_obj_prop(o, "Cleanable")), \
                   lambda r: (not is_of_type(r, "SinkBasin"))
        elif goal_type == "pick_heat_then_place_in_recep":
            return lambda o: (not is_obj_type_in_recep(o, "Microwave")) and \
                             (is_obj_prop(o, "Heatable")), \
                   lambda r: (not is_of_type(r, "Microwave"))
        elif goal_type == "pick_cool_then_place_in_recep":
            return lambda o: (not is_obj_type_in_recep(o, "Fridge")) and \
                             (is_obj_prop(o, "Coolable")), \
                   lambda r: (not is_of_type(r, "Fridge"))
        elif goal_type == "look_at_obj_in_light":
            return lambda o: is_obj_pickupable(o), \
                   lambda r: r
        elif goal_type == "pick_and_place_with_movable_recep":
            return lambda o: is_obj_pickupable(o) and \
                             (not is_movable_recep(o)), \
                   lambda r: is_receptacle(r) and \
                             (not is_movable_recep(r)) and \
                             (not does_any_recep_type_have_obj_of_type(r, self.rand_chosen_val_moveable_recep_class))
        elif goal_type == "pick_heat_and_place_with_movable_recep":
            return lambda o: is_obj_pickupable(o) and \
                             (is_obj_prop(o, "Heatable")) and \
                             (not is_movable_recep(o)), \
                   lambda r: is_receptacle(r) and \
                             (not is_movable_recep(r)) and \
                             (not is_of_type(r, "Microwave")) and \
                             (not does_any_recep_type_have_obj_of_type(r, self.rand_chosen_val_moveable_recep_class))
        else:
            return lambda o: is_obj_pickupable(o), \
                   lambda r: is_receptacle(r)


    def setup_problem(self, seed=None, info=None, scene=None, objs=None):
        '''
        setup goal with sampled objects or with the objects specified
        note: must be used with `initialize_random_scene`
        '''
        self.terminal = False
        print('setup random goal ----------------------------------------------')
        print('seed', seed)
        print('info', info)
        print('--------------------------------------------------------------------\n\n')
        super(TaskGameState, self).setup_problem(seed)
        info, max_num_repeats, remove_prob = self.get_random_task_vals(scene=scene)
        dataset_type, task_row = info

        print('Type:', dataset_type, 'Row: ', task_row, 'Scene', self.scene_name, 'seed', self.scene_seed)
        perform_sanity_check = True

        # pickupable, receptacle, toggleable candidates
        pickupable_objects = [o for o in self.env.last_event.metadata['objects']
                              if (o['pickupable'] and o['objectType'] in constants.OBJECTS)]
        receptacle_objects = [o for o in self.env.last_event.metadata['objects']
                              if (o['receptacle'] and o['objectType'] in constants.RECEPTACLES
                                  and o['objectType'] not in constants.MOVABLE_RECEPTACLES_SET)]
        toggle_objects = [o for o in self.env.last_event.metadata['objects']
                          if o['toggleable'] and not o['isToggled']  # must be one we can turn on.
                          and o['objectType'] in constants.VAL_ACTION_OBJECTS["Toggleable"]]

        if len(pickupable_objects) == 0:
            print("Task Failed - %s" % constants.pddl_goal_type)
            raise Exception("No pickupable objects in the scene")

        # filter based on crit
        obj_crit, _ = self.get_filter_crit(constants.pddl_goal_type)
        pickupable_objects = list(filter(obj_crit, pickupable_objects))
        if len(pickupable_objects) == 0:
            print("Task Failed - %s" % constants.pddl_goal_type)
            raise Exception("No pickupable objects related to the goal '%s'" % constants.pddl_goal_type)

        # choose a pickupable object
        if constants.FORCED_SAMPLING or objs is None:
            self.rand_chosen_object = random.choice(pickupable_objects)
        else:
            chosen_class_available = [obj for obj in pickupable_objects
                                      if obj['objectType'] == constants.OBJ_PARENTS[objs['pickup']]]
            if len(chosen_class_available) == 0:
                print("Task Failed - %s" % constants.pddl_goal_type)
                raise Exception("Couldn't find a valid parent '%s' for pickup object with class '%s'" %
                                 (constants.OBJ_PARENTS[objs['pickup']], objs['pickup']))
            self.rand_chosen_object = random.choice(chosen_class_available)
        self.rand_chosen_object_class = self.rand_chosen_object['objectType']
        self.object_target = constants.OBJECTS.index(self.rand_chosen_object_class)

        # for now, any obj differing from its parent is obtained via slicing, but that may change in the future.
        if constants.OBJ_PARENTS[objs['pickup']] != objs['pickup']:
            requires_slicing = True
        else:
            requires_slicing = False

        # choose a movable receptacle
        if "movable_recep" in constants.pddl_goal_type:
            val_movable_receps = [o for o in self.env.last_event.metadata['objects']
                                  if (o['objectType'] in constants.MOVABLE_RECEPTACLES_SET) and
                                     (self.rand_chosen_object_class in constants.VAL_RECEPTACLE_OBJECTS[o['objectType']])]

            if len(val_movable_receps) == 0:
                print("Task Failed - %s" % constants.pddl_goal_type)
                raise Exception("Couldn't find a valid moveable receptacle for the chosen object class")

            if objs is not None:
                val_movable_receps = [o for o in val_movable_receps if o['objectType'] == objs['mrecep']]

            self.rand_chosen_val_moveable_recep_class = random.choice(val_movable_receps)['objectType']
            self.mrecep_target = constants.OBJECTS.index(self.rand_chosen_val_moveable_recep_class)
        else:
            self.mrecep_target = None

        # if slicing is required, make sure a knife is available in the scene
        if requires_slicing:
            knife_objs = [o for o in self.env.last_event.metadata['objects']
                          if ("Knife" in o['objectType'])]
            if len(knife_objs) == 0:
                print("Task Failed - %s" % constants.pddl_goal_type)
                raise Exception("Couldn't find knife in the scene to cut with")

        if constants.pddl_goal_type == "look_at_obj_in_light":
            # choose a toggleable object
            self.parent_target = None

            if constants.FORCED_SAMPLING or objs is None:
                rand_chosen_toggle_object = random.choice(toggle_objects)
            else:
                toggle_class_available = [obj for obj in toggle_objects
                                          if obj['objectType'] == objs['toggle']]
                if len(toggle_class_available) == 0:
                    print("Task Failed - %s" % constants.pddl_goal_type)
                    raise Exception("Couldn't find a valid toggle object of class '%s'" % objs['toggle'])
                rand_chosen_toggle_object = random.choice(toggle_class_available)

            rand_chosen_toggle_class = rand_chosen_toggle_object['objectType']
            self.toggle_target = constants.OBJECTS.index(rand_chosen_toggle_class)
        else:
            self.toggle_target = None

            # find all valid receptacles in which rand_chosen object or rand chosen moveable receptacle can be placed
            val_receptacle_objects_orig = [r for r in receptacle_objects if (self.rand_chosen_val_moveable_recep_class if self.mrecep_target != None else self.rand_chosen_object_class)
                                      in constants.VAL_RECEPTACLE_OBJECTS[r['objectType']]]
            _, recep_crit = self.get_filter_crit(constants.pddl_goal_type)
            val_receptacle_objects = list(filter(recep_crit, val_receptacle_objects_orig))

            if len(val_receptacle_objects) == 0:
                print("Task Failed - %s" % constants.pddl_goal_type)
                raise Exception("Couldn't find a valid receptacle object according to constraints specified")

            # choose a receptacle object
            if constants.FORCED_SAMPLING or objs is None:
                rand_chosen_receptacle_object = random.choice(val_receptacle_objects)
            else:
                receptacle_class_available = [obj for obj in val_receptacle_objects
                                              if obj['objectType'] == objs['receptacle']]
                if len(receptacle_class_available) == 0:
                    print("Task Failed - %s" % constants.pddl_goal_type)
                    raise Exception("Couldn't find a valid receptacle object of class '%s'" % objs['receptacle'])
                rand_chosen_receptacle_object = random.choice(receptacle_class_available)
            rand_chosen_receptacle_object_class = rand_chosen_receptacle_object['objectType']
            self.parent_target = constants.OBJECTS.index(rand_chosen_receptacle_object_class)

        # pddl_param dict
        constants.data_dict['pddl_params']['object_target'] = constants.OBJECTS[self.object_target] if self.object_target is not None else ""
        constants.data_dict['pddl_params']['object_sliced'] = requires_slicing
        constants.data_dict['pddl_params']['parent_target'] = constants.OBJECTS[self.parent_target] if self.parent_target is not None else ""
        constants.data_dict['pddl_params']['toggle_target'] = constants.OBJECTS[self.toggle_target] if self.toggle_target is not None else ""
        constants.data_dict['pddl_params']['mrecep_target'] = constants.OBJECTS[self.mrecep_target] if self.mrecep_target is not None else ""
        self.task_target = (self.object_target, self.parent_target, self.toggle_target, self.mrecep_target)

        if self.parent_target is None:
            validity_check = True
        else:
            objs = game_util.get_objects_of_type(constants.OBJECTS[self.object_target], self.env.last_event.metadata)
            available_receptacles = game_util.get_objects_of_type(
                constants.OBJECTS[self.parent_target], self.env.last_event.metadata)
            validity_check = (len(objs) > 0 and
                              len([obj for obj in available_receptacles]))

        if perform_sanity_check:
            try:
                assert validity_check, 'Task does not seem valid according to scene metadata'
            except AssertionError:
                raise Exception(str(('Row: ', task_row, 'Scene', self.scene_name,
                                      'seed', self.scene_seed)))

        templated_task_desc = self.get_task_str()
        print('problem id', self.problem_id)
        print('Task:', templated_task_desc)
        constants.data_dict['template']['task_desc'] = templated_task_desc

    def get_setup_info(self, info=None, scene=None):
        return self.get_random_task_vals(scene=scene)

    def reset(self, seed=None, info=None, scene=None, objs=None):
        info = super(TaskGameState, self).reset(seed, info, scene=scene, objs=objs)
        self.receptacle_to_point = None
        self.task_target = None
        self.success = False
        return info

    def step(self, action_or_ind):
        action, _ = self.get_action(action_or_ind)
        super(TaskGameState, self).step(action_or_ind)
        self.success = True

    def get_success(self):
        return self.success
